"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var flags_1 = require("./utils/flags");
var isComponent_1 = require("./utils/isComponent");
var createAssignHelper_1 = require("./utils/createAssignHelper");
var isNullOrUndefined_1 = require("./utils/isNullOrUndefined");
var getName_1 = require("./utils/getName");
var getValue_1 = require("./utils/getValue");
var svgAttributes_1 = require("./utils/svgAttributes");
var isNodeNull_1 = require("./utils/isNodeNull");
var handleWhiteSpace_1 = require("./utils/handleWhiteSpace");
var updateSourceFile_1 = require("./updateSourceFile");
var NULL;
// All special attributes
var PROP_HasKeyedChildren = "$HasKeyedChildren";
var PROP_HasNonKeyedChildren = "$HasNonKeyedChildren";
var PROP_VNODE_CHILDREN = "$HasVNodeChildren";
var PROP_ReCreate = "$ReCreate";
var PROP_ChildFlag = "$ChildFlag";
exports.default = function () {
    return function (context) {
        return function (sourceFile) {
            if (sourceFile.isDeclarationFile) {
                return sourceFile;
            }
            context["createVNode"] = false;
            context["createComponentVNode"] = false;
            context["createTextVNode"] = false;
            context["normalizeProps"] = false;
            var newSourceFile = ts.visitEachChild(sourceFile, visitor, context);
            return updateSourceFile_1.default(newSourceFile, context);
        };
        function visitor(node) {
            switch (node.kind) {
                case ts.SyntaxKind.JsxElement:
                    return createVNode(node, node.children);
                case ts.SyntaxKind.JsxSelfClosingElement:
                    return createVNode(node);
                case ts.SyntaxKind.JsxText:
                    var text = handleWhiteSpace_1.default(node.getFullText());
                    if (text !== "") {
                        return ts.createLiteral(text);
                    }
                    break;
                case ts.SyntaxKind.JsxExpression:
                    if (node.expression) {
                        return ts.visitNode(node.expression, visitor);
                    }
                    break;
                default:
                    return ts.visitEachChild(node, visitor, context);
            }
        }
        function createVNode(node, children) {
            var vType;
            var vProps;
            var vChildren;
            var childrenResults = {};
            var text;
            if (children) {
                var openingElement = node.openingElement;
                vType = getVNodeType(openingElement.tagName);
                vProps = getVNodeProps(openingElement.attributes.properties, vType.isComponent);
                childrenResults = getVNodeChildren(children);
                vChildren = childrenResults.children;
            }
            else {
                vType = getVNodeType(node.tagName);
                vProps = getVNodeProps(node.attributes.properties, vType.isComponent);
            }
            var childFlags = flags_1.ChildFlags.HasInvalidChildren;
            var flags = vType.flags;
            var props = vProps.props[0] || ts.createObjectLiteral();
            var childIndex = -1;
            var i = 0;
            if (vProps.hasReCreateFlag) {
                flags = flags | flags_1.VNodeFlags.ReCreate;
            }
            if (vType.isComponent) {
                if (vChildren) {
                    if (!(vChildren.kind === ts.SyntaxKind.ArrayLiteralExpression &&
                        vChildren.elements.length === 0)) {
                        // Remove children from props, if it exists
                        for (i = 0; i < props.properties.length; i++) {
                            if (props.properties[i] &&
                                props.properties[i].name.text === "children") {
                                childIndex = i;
                                break;
                            }
                        }
                        if (childIndex !== -1) {
                            props.properties.splice(childIndex, 1); // Remove prop children
                        }
                        props.properties.push(ts.createPropertyAssignment(getName_1.default("children"), vChildren));
                        vProps.props[0] = props;
                    }
                    vChildren = NULL;
                }
            }
            else {
                if (((vChildren &&
                    vChildren.kind === ts.SyntaxKind.ArrayLiteralExpression) ||
                    !vChildren) &&
                    vProps.propChildren) {
                    if (vProps.propChildren.kind === ts.SyntaxKind.StringLiteral) {
                        text = handleWhiteSpace_1.default(vProps.propChildren.text);
                        if (text !== "") {
                            childrenResults.foundText = true;
                            childrenResults.hasSingleChild = true;
                            vChildren = ts.createLiteral(text);
                        }
                    }
                    else if (vProps.propChildren.kind === ts.SyntaxKind.JsxExpression) {
                        if (vProps.propChildren.expression.kind === ts.SyntaxKind.NullKeyword) {
                            vChildren = NULL;
                            childFlags = flags_1.ChildFlags.HasInvalidChildren;
                        }
                        else {
                            vChildren = createVNode(vProps.propChildren.expression, vProps.propChildren.expression.children);
                            childFlags = flags_1.ChildFlags.HasVNodeChildren;
                        }
                    }
                    else {
                        vChildren = NULL;
                        childFlags = flags_1.ChildFlags.HasInvalidChildren;
                    }
                }
                if ((childrenResults && !childrenResults.requiresNormalization) ||
                    vProps.childrenKnown) {
                    if (vProps.hasKeyedChildren || childrenResults.parentCanBeKeyed) {
                        childFlags = flags_1.ChildFlags.HasKeyedChildren;
                    }
                    else if (vProps.hasNonKeyedChildren ||
                        childrenResults.parentCanBeNonKeyed) {
                        childFlags = flags_1.ChildFlags.HasNonKeyedChildren;
                    }
                    else if (childrenResults.hasSingleChild) {
                        childFlags = flags_1.ChildFlags.HasVNodeChildren;
                    }
                }
                else {
                    if (vProps.hasKeyedChildren) {
                        childFlags = flags_1.ChildFlags.HasKeyedChildren;
                    }
                    else if (vProps.hasNonKeyedChildren) {
                        childFlags = flags_1.ChildFlags.HasNonKeyedChildren;
                    }
                }
                // Remove children from props, if it exists
                childIndex = -1;
                for (i = 0; i < props.properties.length; i++) {
                    if (props.properties[i].name &&
                        props.properties[i].name.text === "children") {
                        childIndex = i;
                        break;
                    }
                }
                if (childIndex !== -1) {
                    props.properties.splice(childIndex, 1); // Remove prop children
                }
            }
            if (vChildren && vChildren !== NULL) {
                if (childrenResults.foundText) {
                    if (childrenResults.hasSingleChild) {
                        context["createTextVNode"] = true;
                        vChildren = ts.createCall(ts.createIdentifier("createTextVNode"), [], [vChildren]);
                    }
                    else if (vChildren.elements) {
                        addCreateTextVNodeCalls(vChildren);
                    }
                }
            }
            var willNormalizeChildren = !vType.isComponent &&
                childrenResults &&
                childrenResults.requiresNormalization &&
                !vProps.childrenKnown;
            if (vProps.childFlags) {
                // If $ChildFlag is provided it is runtime dependant
                childFlags = vProps.childFlags;
            }
            else {
                childFlags = willNormalizeChildren
                    ? flags_1.ChildFlags.UnknownChildren
                    : childFlags;
            }
            // Delete empty objects
            if (vProps.props.length === 1 &&
                vProps.props[0] &&
                !vProps.props[0].properties.length) {
                vProps.props.splice(0, 1);
            }
            var createVNodeCall = ts.createCall(vType.isComponent
                ? ts.createIdentifier("createComponentVNode")
                : ts.createIdentifier("createVNode"), [], vType.isComponent
                ? createComponentVNodeArgs(flags, vType.type, vProps.props, vProps.key, vProps.ref)
                : createVNodeArgs(flags, vType.type, vProps.className, vChildren, childFlags, vProps.props, vProps.key, vProps.ref, context));
            if (vType.isComponent) {
                context["createComponentVNode"] = true;
            }
            else {
                context["createVNode"] = true;
            }
            // NormalizeProps will normalizeChildren too
            if (vProps.needsNormalization) {
                context["normalizeProps"] = true;
                createVNodeCall = ts.createCall(ts.createIdentifier("normalizeProps"), [], [createVNodeCall]);
            }
            return createVNodeCall;
        }
        function addCreateTextVNodeCalls(vChildren) {
            // When normalization is not needed we need to manually compile text into vNodes
            for (var j = 0; j < vChildren.elements.length; j++) {
                var aChild = vChildren.elements[j];
                if (aChild.kind === ts.SyntaxKind.StringLiteral) {
                    context["createTextVNode"] = true;
                    vChildren.elements[j] = ts.createCall(ts.createIdentifier("createTextVNode"), [], [aChild]);
                }
            }
        }
        function getVNodeType(type) {
            var nodeKind = type.kind;
            var component = false;
            var flags;
            if (nodeKind === ts.SyntaxKind.Identifier || nodeKind === ts.SyntaxKind.PropertyAccessExpression) {
                if (isComponent_1.default(type.text || type.expression.text)) {
                    component = true;
                    flags = flags_1.VNodeFlags.ComponentUnknown;
                }
                else {
                    var tag = type.text;
                    type = ts.createLiteral(tag);
                    switch (tag) {
                        case "svg":
                            flags = flags_1.VNodeFlags.SvgElement;
                            break;
                        case "input":
                            flags = flags_1.VNodeFlags.InputElement;
                            break;
                        case "textarea":
                            flags = flags_1.VNodeFlags.TextareaElement;
                            break;
                        case "select":
                            flags = flags_1.VNodeFlags.SelectElement;
                            break;
                        default:
                            flags = flags_1.VNodeFlags.HtmlElement;
                    }
                }
            }
            return {
                type: type,
                isComponent: component,
                flags: flags
            };
        }
        function getVNodeProps(astProps, isComponent) {
            var props = [];
            var key = null;
            var ref = null;
            var className = null;
            var hasKeyedChildren = false;
            var hasNonKeyedChildren = false;
            var childrenKnown = false;
            var needsNormalization = false;
            var hasReCreateFlag = false;
            var propChildren = null;
            var childFlags = null;
            var assignArgs = [];
            for (var i = 0; i < astProps.length; i++) {
                var astProp = astProps[i];
                var initializer = astProp.initializer;
                if (astProp.kind === ts.SyntaxKind.JsxSpreadAttribute) {
                    needsNormalization = true;
                    assignArgs = [
                        ts.createObjectLiteral(),
                        astProp.expression
                    ];
                }
                else {
                    var propName = astProp.name.text;
                    if (!isComponent &&
                        (propName === "className" || propName === "class")) {
                        className = getValue_1.default(initializer, visitor);
                    }
                    else if (!isComponent && propName === "htmlFor") {
                        props.push(ts.createPropertyAssignment(getName_1.default("for"), getValue_1.default(initializer, visitor)));
                    }
                    else if (!isComponent && propName === "onDoubleClick") {
                        props.push(ts.createPropertyAssignment(getName_1.default("onDblClick"), getValue_1.default(initializer, visitor)));
                    }
                    else if (propName.substr(0, 11) === "onComponent" && isComponent) {
                        if (!ref) {
                            ref = ts.createObjectLiteral([]);
                        }
                        ref.properties.push(ts.createPropertyAssignment(getName_1.default(propName), getValue_1.default(initializer, visitor)));
                    }
                    else if (!isComponent && propName in svgAttributes_1.default) {
                        // React compatibility for SVG Attributes
                        props.push(ts.createPropertyAssignment(getName_1.default(svgAttributes_1.default[propName]), getValue_1.default(initializer, visitor)));
                    }
                    else {
                        switch (propName) {
                            case "noNormalize":
                            case "$NoNormalize":
                                throw "Inferno JSX plugin:\n" +
                                    propName +
                                    " is deprecated use: $HasVNodeChildren, or if children shape is dynamic you can use: $ChildFlag={expression} see inferno package:inferno-vnode-flags (ChildFlags) for possible values";
                            case "hasKeyedChildren":
                            case "hasNonKeyedChildren":
                                throw "Inferno JSX plugin:\n" +
                                    propName +
                                    " is deprecated use: " +
                                    "$" +
                                    propName.charAt(0).toUpperCase() +
                                    propName.slice(1);
                            case PROP_ChildFlag:
                                childrenKnown = true;
                                childFlags = getValue_1.default(initializer, visitor);
                                break;
                            case PROP_VNODE_CHILDREN:
                                childrenKnown = true;
                                break;
                            case PROP_HasNonKeyedChildren:
                                hasNonKeyedChildren = true;
                                childrenKnown = true;
                                break;
                            case PROP_HasKeyedChildren:
                                hasKeyedChildren = true;
                                childrenKnown = true;
                                break;
                            case "ref":
                                ref = getValue_1.default(initializer, visitor);
                                break;
                            case "key":
                                key = getValue_1.default(initializer, visitor);
                                break;
                            case PROP_ReCreate:
                                hasReCreateFlag = true;
                                break;
                            default:
                                if (propName === "children") {
                                    propChildren = astProp.initializer;
                                }
                                props.push(ts.createPropertyAssignment(getName_1.default(propName), initializer
                                    ? getValue_1.default(initializer, visitor)
                                    : ts.createTrue()));
                        }
                    }
                }
            }
            if (props.length)
                assignArgs.push(ts.createObjectLiteral(props));
            return {
                props: assignArgs,
                key: isNullOrUndefined_1.default(key) ? NULL : key,
                ref: isNullOrUndefined_1.default(ref) ? NULL : ref,
                hasKeyedChildren: hasKeyedChildren,
                hasNonKeyedChildren: hasNonKeyedChildren,
                propChildren: propChildren,
                childrenKnown: childrenKnown,
                className: isNullOrUndefined_1.default(className) ? NULL : className,
                childFlags: childFlags,
                hasReCreateFlag: hasReCreateFlag,
                needsNormalization: needsNormalization
            };
        }
        function getVNodeChildren(astChildren) {
            ts.SyntaxKind.NumericLiteral;
            var children = [];
            var parentCanBeKeyed = false;
            var requiresNormalization = false;
            var foundText = false;
            for (var i = 0; i < astChildren.length; i++) {
                var child = astChildren[i];
                var vNode = visitor(child);
                if (child.kind === ts.SyntaxKind.JsxExpression) {
                    requiresNormalization = true;
                }
                else if (child.kind === ts.SyntaxKind.JsxText &&
                    handleWhiteSpace_1.default(child.getText()) !== "") {
                    foundText = true;
                }
                if (!isNullOrUndefined_1.default(vNode)) {
                    children.push(vNode);
                    /*
                      * Loop direct children to check if they have key property set
                      * If they do, flag parent as hasKeyedChildren to increase runtime performance of Inferno
                      * When key already found within one of its children, they must all be keyed
                      */
                    if (parentCanBeKeyed === false && child.openingElement) {
                        var astProps = child.openingElement.attributes;
                        var len = astProps.length;
                        while (parentCanBeKeyed === false && len-- > 0) {
                            var prop = astProps[len];
                            if (prop.name && prop.name.name === "key") {
                                parentCanBeKeyed = true;
                            }
                        }
                    }
                }
            }
            // Fix: When there is single child parent cant be keyed either, its faster to use patch than patchKeyed routine in that case
            var hasSingleChild = children.length === 1;
            return {
                parentCanBeKeyed: hasSingleChild === false && parentCanBeKeyed,
                children: hasSingleChild
                    ? children[0]
                    : ts.createArrayLiteral(children),
                foundText: foundText,
                parentCanBeNonKeyed: !hasSingleChild &&
                    !parentCanBeKeyed &&
                    !requiresNormalization &&
                    astChildren.length > 1,
                requiresNormalization: requiresNormalization,
                hasSingleChild: hasSingleChild
            };
        }
        function createComponentVNodeArgs(flags, type, props, key, ref) {
            var args = [];
            var hasProps = props.length > 0;
            var hasKey = !isNodeNull_1.default(key);
            var hasRef = !isNodeNull_1.default(ref);
            args.push(ts.createNumericLiteral(flags + ""));
            args.push(type);
            if (hasProps) {
                props.length === 1
                    ? args.push(props[0])
                    : args.push(createAssignHelper_1.default(context, props));
            }
            else if (hasKey || hasRef) {
                args.push(ts.createNull());
            }
            if (hasKey) {
                args.push(key);
            }
            else if (hasRef) {
                args.push(ts.createNull());
            }
            if (hasRef) {
                args.push(ref);
            }
            return args;
        }
        function createVNodeArgs(flags, type, className, children, childFlags, props, key, ref, context) {
            var args = [];
            var hasClassName = !isNodeNull_1.default(className);
            var hasChildren = !isNodeNull_1.default(children);
            var hasChildFlags = childFlags !== flags_1.ChildFlags.HasInvalidChildren;
            var hasProps = props.length > 0;
            var hasKey = !isNodeNull_1.default(key);
            var hasRef = !isNodeNull_1.default(ref);
            args.push(ts.createNumericLiteral(flags + ""));
            args.push(type);
            if (hasClassName) {
                args.push(className);
            }
            else if (hasChildren || hasChildFlags || hasProps || hasKey || hasRef) {
                args.push(ts.createNull());
            }
            if (hasChildren) {
                args.push(children);
            }
            else if (hasChildFlags || hasProps || hasKey || hasRef) {
                args.push(ts.createNull());
            }
            if (hasChildFlags) {
                args.push(typeof childFlags === "number"
                    ? ts.createNumericLiteral(childFlags + "")
                    : childFlags);
            }
            else if (hasProps || hasKey || hasRef) {
                args.push(ts.createNumericLiteral(flags_1.ChildFlags.HasInvalidChildren + ""));
            }
            if (hasProps) {
                props.length === 1
                    ? args.push(props[0])
                    : args.push(createAssignHelper_1.default(context, props));
            }
            else if (hasKey || hasRef) {
                args.push(ts.createNull());
            }
            if (hasKey) {
                args.push(key);
            }
            else if (hasRef) {
                args.push(ts.createNull());
            }
            if (hasRef) {
                args.push(ref);
            }
            return args;
        }
    };
};
